#
# Autogenerated by Thrift Compiler (0.13.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class JobState(object):
    SUBMITTED = 1
    SCHEDULED = 2
    DISPATCHED = 3
    RUNNING = 4
    FINISHED = 5
    FAILED = 6

    _VALUES_TO_NAMES = {
        1: "SUBMITTED",
        2: "SCHEDULED",
        3: "DISPATCHED",
        4: "RUNNING",
        5: "FINISHED",
        6: "FAILED",
    }

    _NAMES_TO_VALUES = {
        "SUBMITTED": 1,
        "SCHEDULED": 2,
        "DISPATCHED": 3,
        "RUNNING": 4,
        "FINISHED": 5,
        "FAILED": 6,
    }


class JobType(object):
    COPY = 1
    UPLOAD = 2
    MOCK = 3

    _VALUES_TO_NAMES = {
        1: "COPY",
        2: "UPLOAD",
        3: "MOCK",
    }

    _NAMES_TO_VALUES = {
        "COPY": 1,
        "UPLOAD": 2,
        "MOCK": 3,
    }


class JobStatus(object):
    """
    Attributes:
     - id
     - state
     - submission_time
     - start_time
     - finish_time
     - execution_time

    """


    def __init__(self, id=None, state=None, submission_time=None, start_time=None, finish_time=None, execution_time=None,):
        self.id = id
        self.state = state
        self.submission_time = submission_time
        self.start_time = start_time
        self.finish_time = finish_time
        self.execution_time = execution_time

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.id = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I32:
                    self.state = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.STRING:
                    self.submission_time = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.STRING:
                    self.start_time = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.STRING:
                    self.finish_time = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 6:
                if ftype == TType.I32:
                    self.execution_time = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('JobStatus')
        if self.id is not None:
            oprot.writeFieldBegin('id', TType.STRING, 1)
            oprot.writeString(self.id.encode('utf-8') if sys.version_info[0] == 2 else self.id)
            oprot.writeFieldEnd()
        if self.state is not None:
            oprot.writeFieldBegin('state', TType.I32, 2)
            oprot.writeI32(self.state)
            oprot.writeFieldEnd()
        if self.submission_time is not None:
            oprot.writeFieldBegin('submission_time', TType.STRING, 3)
            oprot.writeString(self.submission_time.encode('utf-8') if sys.version_info[0] == 2 else self.submission_time)
            oprot.writeFieldEnd()
        if self.start_time is not None:
            oprot.writeFieldBegin('start_time', TType.STRING, 4)
            oprot.writeString(self.start_time.encode('utf-8') if sys.version_info[0] == 2 else self.start_time)
            oprot.writeFieldEnd()
        if self.finish_time is not None:
            oprot.writeFieldBegin('finish_time', TType.STRING, 5)
            oprot.writeString(self.finish_time.encode('utf-8') if sys.version_info[0] == 2 else self.finish_time)
            oprot.writeFieldEnd()
        if self.execution_time is not None:
            oprot.writeFieldBegin('execution_time', TType.I32, 6)
            oprot.writeI32(self.execution_time)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Job(object):
    """
    Attributes:
     - type
     - params

    """


    def __init__(self, type=None, params=None,):
        self.type = type
        self.params = params

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.type = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.MAP:
                    self.params = {}
                    (_ktype1, _vtype2, _size0) = iprot.readMapBegin()
                    for _i4 in range(_size0):
                        _key5 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        _val6 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.params[_key5] = _val6
                    iprot.readMapEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Job')
        if self.type is not None:
            oprot.writeFieldBegin('type', TType.I32, 1)
            oprot.writeI32(self.type)
            oprot.writeFieldEnd()
        if self.params is not None:
            oprot.writeFieldBegin('params', TType.MAP, 2)
            oprot.writeMapBegin(TType.STRING, TType.STRING, len(self.params))
            for kiter7, viter8 in self.params.items():
                oprot.writeString(kiter7.encode('utf-8') if sys.version_info[0] == 2 else kiter7)
                oprot.writeString(viter8.encode('utf-8') if sys.version_info[0] == 2 else viter8)
            oprot.writeMapEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Worker(object):
    """
    Attributes:
     - wid
     - address
     - port

    """


    def __init__(self, wid=None, address=None, port=None,):
        self.wid = wid
        self.address = address
        self.port = port

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.wid = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.address = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I32:
                    self.port = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Worker')
        if self.wid is not None:
            oprot.writeFieldBegin('wid', TType.STRING, 1)
            oprot.writeString(self.wid.encode('utf-8') if sys.version_info[0] == 2 else self.wid)
            oprot.writeFieldEnd()
        if self.address is not None:
            oprot.writeFieldBegin('address', TType.STRING, 2)
            oprot.writeString(self.address.encode('utf-8') if sys.version_info[0] == 2 else self.address)
            oprot.writeFieldEnd()
        if self.port is not None:
            oprot.writeFieldBegin('port', TType.I32, 3)
            oprot.writeI32(self.port)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.wid is None:
            raise TProtocolException(message='Required field wid is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class FileSystem(object):
    """
    Attributes:
     - name
     - parameters

    """


    def __init__(self, name=None, parameters=None,):
        self.name = name
        self.parameters = parameters

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.name = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.MAP:
                    self.parameters = {}
                    (_ktype10, _vtype11, _size9) = iprot.readMapBegin()
                    for _i13 in range(_size9):
                        _key14 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        _val15 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.parameters[_key14] = _val15
                    iprot.readMapEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('FileSystem')
        if self.name is not None:
            oprot.writeFieldBegin('name', TType.STRING, 1)
            oprot.writeString(self.name.encode('utf-8') if sys.version_info[0] == 2 else self.name)
            oprot.writeFieldEnd()
        if self.parameters is not None:
            oprot.writeFieldBegin('parameters', TType.MAP, 2)
            oprot.writeMapBegin(TType.STRING, TType.STRING, len(self.parameters))
            for kiter16, viter17 in self.parameters.items():
                oprot.writeString(kiter16.encode('utf-8') if sys.version_info[0] == 2 else kiter16)
                oprot.writeString(viter17.encode('utf-8') if sys.version_info[0] == 2 else viter17)
            oprot.writeMapEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.name is None:
            raise TProtocolException(message='Required field name is unset!')
        if self.parameters is None:
            raise TProtocolException(message='Required field parameters is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class JobInitializationException(TException):
    """
    Attributes:
     - message

    """


    def __init__(self, message=None,):
        self.message = message

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.message = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('JobInitializationException')
        if self.message is not None:
            oprot.writeFieldBegin('message', TType.STRING, 1)
            oprot.writeString(self.message.encode('utf-8') if sys.version_info[0] == 2 else self.message)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __str__(self):
        return repr(self)

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class JobNotFoundException(TException):
    """
    Attributes:
     - message

    """


    def __init__(self, message=None,):
        self.message = message

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.message = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('JobNotFoundException')
        if self.message is not None:
            oprot.writeFieldBegin('message', TType.STRING, 1)
            oprot.writeString(self.message.encode('utf-8') if sys.version_info[0] == 2 else self.message)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __str__(self):
        return repr(self)

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(JobStatus)
JobStatus.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'id', 'UTF8', None, ),  # 1
    (2, TType.I32, 'state', None, None, ),  # 2
    (3, TType.STRING, 'submission_time', 'UTF8', None, ),  # 3
    (4, TType.STRING, 'start_time', 'UTF8', None, ),  # 4
    (5, TType.STRING, 'finish_time', 'UTF8', None, ),  # 5
    (6, TType.I32, 'execution_time', None, None, ),  # 6
)
all_structs.append(Job)
Job.thrift_spec = (
    None,  # 0
    (1, TType.I32, 'type', None, None, ),  # 1
    (2, TType.MAP, 'params', (TType.STRING, 'UTF8', TType.STRING, 'UTF8', False), None, ),  # 2
)
all_structs.append(Worker)
Worker.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'wid', 'UTF8', None, ),  # 1
    (2, TType.STRING, 'address', 'UTF8', None, ),  # 2
    (3, TType.I32, 'port', None, None, ),  # 3
)
all_structs.append(FileSystem)
FileSystem.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'name', 'UTF8', None, ),  # 1
    (2, TType.MAP, 'parameters', (TType.STRING, 'UTF8', TType.STRING, 'UTF8', False), None, ),  # 2
)
all_structs.append(JobInitializationException)
JobInitializationException.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'message', 'UTF8', None, ),  # 1
)
all_structs.append(JobNotFoundException)
JobNotFoundException.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'message', 'UTF8', None, ),  # 1
)
fix_spec(all_structs)
del all_structs



import abc
import logging as lg
import time
from threading import Thread

import six

_logger = lg.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class Dispatcher(object):
    """
    The Dispatcher distribute scheduled tasks between workers.

    Initially tasks are queued when submitted to the dispatcher
    then the dispatch thread continuously poll tasks and select
    the appropriate worker for execution.
    """

    def __init__(self, cqueue, workers_manager, callback=None):
        self.workers_manager = workers_manager
        self.callback = callback
        self.cqueue = cqueue
        self.stopped = False

    def _run(self):
        task_exec = None
        while True:
            if not self.cqueue.empty():
                worker = self.next_worker()
                if worker is None:
                    _logger.warn("Failed to dispatch tasks : no workers found !")
                    time.sleep(2)
                else:
                    task_exec = self.cqueue.get() if task_exec is None else task_exec
                    _logger.debug(
                        "Dispatching next task [ %s ] for execution.", task_exec.tid
                    )
                    if self.callback is not None:
                        self.callback(task_exec.tid, worker.wid)
                    try:
                        worker.submit(task_exec)
                        task_exec = None
                    except Exception:
                        _logger.exception(
                            "Exception submitting task [ %s ] to worker [ %s ]",  # NOQA
                            task_exec.tid,
                            worker.wid,
                        )

            else:
                _logger.debug("No new tasks to dispatch, sleeping for %s seconds...", 2)
                time.sleep(2)

            if self.stopped and self.cqueue.empty():
                _logger.debug(
                    "No new tasks to dispatch, terminating dispatcher thread."
                )
                return

    @abc.abstractmethod
    def next_worker(self):
        return

    def start(self):
        _logger.info("Stating tanit dispatcher.")
        self.daemon = Thread(target=self._run, args=())
        self.daemon.setDaemon(True)
        self.daemon.start()
        _logger.info("Tanit dispatcher started.")

    def stop(self):
        _logger.info("Stopping tanit dispatcher.")
        self.stopped = True
        self.daemon.join()
        _logger.info("tanit dispatcher Stopped.")


class FairDispatcher(Dispatcher):
    def next_worker(self):
        live_workers = self.workers_manager.list_live_workers()

        if len(live_workers) == 0:
            return None

        best_status = None
        best_worker = None

        for worker in live_workers:
            try:
                status = worker.status()
            except Exception:
                _logger.exception(
                    "Exception while fetching worker [ %s ] status", worker.wid
                )
                continue
            if best_status is None:
                best_status = status
                best_worker = worker
            elif status.pending < best_status.pending:
                best_worker = worker
                best_status = status
                continue
            elif status.pending == best_status.pending:
                if status.available > best_status.available:
                    best_worker = worker
                    best_status = status
                    continue
        return best_worker



import pytest

from tanit.common.model.execution_type import ExecutionType
from tanit.common.model.job import Job

from ..unit.utils.tutils import wait_until


@pytest.mark.usefixtures("master_server")
class TestMasterStandalone:
    def test_server_up(self, user_client):
        assert len(user_client.list_jobs()) >= 0

    def test_register_local_worker(self, worker_client):
        worker_client.register_worker("local-worker", None, None)
        assert len(worker_client.list_workers()) == 1

    def test_submit_job(self, user_client):
        jid = user_client.submit_job(Job(ExecutionType.MOCK, {"num_tasks": "10"}))
        # wait for the job to finish
        assert wait_until(
            lambda i: user_client.job_status(i).state == "FINISHED",
            20,
            0.5,
            jid,
        )

    def test_list_jobs(self, user_client):
        assert len(user_client.list_jobs()) >= 1

    def test_unregister_local_worker(self, worker_client):
        worker_client.unregister_worker("local-worker", None, None)
        assert len(worker_client.list_workers()) == 0



import time


def wait_until(condition, timeout, period=0.25, *args, **kwargs):
    must_end = time.time() + timeout
    while time.time() < must_end:
        if condition(*args, **kwargs):
            return True
        time.sleep(period)
    return False



class Job(object):
    def __init__(self, etype, params):
        self.etype = etype
        self.params = params


class JobStatus(object):
    def __init__(
        self, jid, state, submission_time, start_time, finish_time, execution_time
    ):
        self.jid = jid
        self.state = state
        self.submission_time = submission_time
        self.start_time = start_time
        self.finish_time = finish_time
        self.execution_time = execution_time

    def __str__(self):
        return (
            "JobStatus {"
            "id: %s,"
            "state: %s,"
            "submission_time: %s,"
            "start_time: %s,"
            "finish_time: %s,"
            "execution_time: %s"
            "}"
            % (
                self.jid,
                self.state,
                self.submission_time,
                self.start_time,
                self.finish_time,
                self.execution_time,
            )
        )



from tanit.common.model.worker import WorkerStatus
from tanit.master.core.worker.worker import WorkerIFace
from tanit.master.core.worker.worker_factory import WorkerFactory


class MockWorker(WorkerIFace):
    def __init__(self, wid, cores=10):
        super(MockWorker, self).__init__(wid, None, None)

        self.tasks = []
        self.cores = cores

    def start(self):
        super(MockWorker, self).start()

    def stop(self):
        super(MockWorker, self).stop()

    def submit(self, task):
        self.tasks.append(task)

    def register_filesystem(self, name, filesystem):
        pass

    def status(self):
        return WorkerStatus(
            self.wid,
            len(self.tasks),
            0 if len(self.tasks) < self.cores else len(self.tasks) - self.cores,
            0 if self.cores < len(self.tasks) else self.cores - len(self.tasks),
        )


class MockWorkerFactory(WorkerFactory):
    def create_worker(self, worker):
        return MockWorker(worker.wid, worker.cores)



import os
import random


class BaseFilesystemTest:
    """Test the filesystem interface.

    The same set of tests need to run successfully on
    all the filesystem implementations regardless of their type,
    This verify that all filesystem implementations respect
    the same interface and can be used transparently regardless
    of the underlying implementation.
    """

    def test_mkdir(self, filesystem, test_data):
        for i in range(5):
            d = os.path.join(test_data, "ktd-%s" % i)
            filesystem.mkdir(d)
        assert len(filesystem.list(test_data)) == 5

    def test_create_file(self, filesystem, test_data):
        for f in filesystem.list(test_data):
            for i in range(5):
                with filesystem.open(
                    os.path.join(test_data, f, "ktf-%s.txt" % i), mode="wb"
                ) as tfile:
                    tfile.write(bytearray(random.getrandbits(8) for _ in range(1024)))
            assert len(filesystem.list(os.path.join(test_data, f))) == 5

    def test_list(self, filesystem, test_data):
        assert len(filesystem.list(test_data)) == 5

    def test_list_glob(self, filesystem, test_data):
        assert len(filesystem.list(os.path.join(test_data, "ktd-*"), glob=True)) == 5

    def test_walk(self, filesystem, test_data):
        for dpath, dnames, fnames in filesystem.walk(test_data):
            assert dpath in [test_data] + [
                os.path.join(test_data, "ktd-%s" % i) for i in range(5)
            ]
            if dpath == test_data:
                assert len(dnames) == 5
                assert len(fnames) == 0
            else:
                assert len(dnames) == 0
                assert len(fnames) == 5

    def test_status(self, filesystem, test_data):
        assert (
            filesystem.status(os.path.join(test_data, "ktd-1"))["type"] == "DIRECTORY"
        )
        assert (
            filesystem.status(os.path.join(test_data, "ktd-1", "ktf-1.txt"))["type"]
            == "FILE"
        )
        assert filesystem.status(os.path.join(test_data, "ktd-1", "ktf-1.txt"))[
            "length"
        ] in ["1024", "1026"]

    def test_content(self, filesystem, test_data):
        assert filesystem.content(os.path.join(test_data))["fileCount"] == "25"
        assert filesystem.content(os.path.join(test_data))["directoryCount"] == "5"

    def test_copy(self, filesystem, test_data):
        test_src_dir = os.path.join(test_data, "ktd-1")
        test_rename_dir = os.path.join(test_data, "ktd-1-copy")

        filesystem.copy(test_src_dir, test_rename_dir)
        assert filesystem.status(test_rename_dir)["type"] == "DIRECTORY"

        src_content = filesystem.content(test_src_dir)
        dst_content = filesystem.content(test_rename_dir)
        assert src_content["length"] == dst_content["length"]
        assert src_content["fileCount"] == dst_content["fileCount"]
        assert src_content["directoryCount"] == dst_content["directoryCount"]

    def test_rename(self, filesystem, test_data):
        test_src_dir = os.path.join(test_data, "ktd-1-copy")
        test_dst_dir = os.path.join(test_data, "ktd-1-renamed")

        src_content = filesystem.content(test_src_dir)
        filesystem.rename(test_src_dir, test_dst_dir)
        assert filesystem.exists(test_dst_dir)
        assert not filesystem.exists(test_src_dir)

        dst_content = filesystem.content(test_dst_dir)
        assert src_content["length"] == dst_content["length"]
        assert src_content["fileCount"] == dst_content["fileCount"]
        assert src_content["directoryCount"] == dst_content["directoryCount"]

    def test_delete(self, filesystem, test_data):
        test_delete_dir = os.path.join(test_data, "ktd-1-renamed")
        filesystem.delete(test_delete_dir, recursive=True)
        assert not filesystem.exists(test_delete_dir)

    def test_write_read(self, filesystem, test_data):
        test_writer_file = os.path.join(test_data, "ktf-write")
        wbts = "This is a test generated file".encode("utf-8")
        with filesystem.write(test_writer_file) as writer:
            writer.write(wbts)
        assert filesystem.status(test_writer_file)["type"] == "FILE"
        assert filesystem.status(test_writer_file)["length"] == str(len(wbts))

        with filesystem.read(test_writer_file) as reader:
            rbts = reader.read()
        assert rbts == wbts



import grp
import io
import logging as lg
import os
import pwd
import shutil
import stat

from ...common.utils.glob import iglob
from ..filesystem import IFileSystem
from ..ioutils import FileSystemError

_logger = lg.getLogger(__name__)


class LocalFileSystem(IFileSystem):
    def resolve_path(self, path):
        return os.path.normpath(os.path.abspath(path))

    def _status(self, path, strict=True):
        def _get_type(p):
            if os.path.isfile(p):
                return "FILE"
            elif os.path.isdir(p):
                return "DIRECTORY"
            elif os.path.islink(p):
                return "LINK"
            else:
                return ""

        _logger.debug("Fetching status for %r.", path)
        rpath = self.resolve_path(path)
        if not os.path.exists(rpath):
            if not strict:
                return None
            else:
                raise FileSystemError("%r does not exist.", rpath)
        else:
            path_stat = os.stat(rpath)
            return {
                "owner": str(pwd.getpwuid(path_stat.st_uid)[0]),
                "childrenNum": str(
                    len(os.listdir(rpath)) if os.path.isdir(rpath) else 0
                ),
                "accessTime": str(int(path_stat.st_atime * 1000)),
                "fileId": str(path_stat.st_ino),
                "permission": str(oct(stat.S_IMODE(path_stat.st_mode))),
                "length": str(path_stat.st_size if os.path.isfile(rpath) else 0),
                "type": str(_get_type(rpath)),
                "group": str(grp.getgrgid(path_stat.st_gid)[0]),
                "modificationTime": str(int(path_stat.st_mtime * 1000)),
            }

    def _list(self, path, status=False, glob=False):
        if not glob:
            files = os.listdir(path)
            if status:
                return [(f, self.status(os.path.join(path, f))) for f in files]
            else:
                return files
        else:
            files = [i_file for i_file in iglob(self, path)]
            if status:
                return [(f, self.status(f)) for f in files]
            else:
                return files

    def _delete(self, path, recursive=False):
        if os.path.isfile(path):
            os.remove(path)
        else:
            if len(os.listdir(path)) == 0:
                os.rmdir(path)
            else:
                if recursive:
                    shutil.rmtree(path)
                else:
                    raise FileSystemError("Directory %r is not empty.", path)

    def _copy(self, src_path, dst_path):
        shutil.copy(src_path, dst_path)

    def rename(self, src_path, dst_path):
        rsrc_path = self.resolve_path(src_path)
        rdst_path = self.resolve_path(dst_path)
        if not os.path.exists(rsrc_path):
            raise FileSystemError("%r does not exist.", rsrc_path)
        if os.path.exists(rdst_path):
            raise FileSystemError("%r exists.", rdst_path)
        os.rename(rsrc_path, rdst_path)

    def _set_owner(self, path, owner=None, group=None):
        uid = pwd.getpwnam(owner).pw_uid if owner is not None else -1
        gid = grp.getgrnam(group).gr_gid if group is not None else -1
        os.chown(path, uid, gid)

    def _set_permission(self, path, permission):
        os.chmod(path, int(permission, 8))

    def _mkdir(self, path, permission=None):
        os.makedirs(path)
        if permission:
            os.chmod(path, int(permission, 8))

    def _open(self, path, mode, buffer_size=0, encoding=None, **kwargs):
        file = open(path, mode=mode, buffering=buffer_size)
        if encoding:
            return io.TextIOWrapper(file, encoding=encoding)
        else:
            return file



import logging as lg

_logger = lg.getLogger(__name__)


class TanitError(Exception):
    """
    Base error class.

     :param message: Error message.
     :param args: optional Message formatting arguments.
    """

    def __init__(self, message, *args):
        super(TanitError, self).__init__(message % args if args else message)



import logging as lg

from ...common.model.worker import Worker
from ..core.master import Master

_logger = lg.getLogger(__name__)


class StandaloneMaster(Master):
    def start(self):
        super(StandaloneMaster, self).start()
        _logger.info("Registering local Worker.")
        self.workers_manager.register_worker(Worker("local-worker", None, None))
        # register the local file system
        self.register_filesystem("local", {"type": "local"})
        _logger.info("Local worker Registered.")

    def register_worker(self, worker):
        raise UnsupportedOperationException(
            "Standalone master does not support registering external workers."
        )


class UnsupportedOperationException(Exception):
    pass



import pytest
from six.moves.queue import Queue

from tanit.common.model.execution_type import ExecutionType
from tanit.common.model.job import Job
from tanit.common.model.worker import Worker
from tanit.master.core.dispatcher import FairDispatcher
from tanit.master.core.execution.job_factory import JobFactory
from tanit.master.core.worker.worker_manager import WorkerManager

from ...utils.tutils import wait_until
from .worker.mock_worker import MockWorkerFactory

job_factory = JobFactory()


def mock_job_exec(num_tasks):
    job = job_factory.create_job(Job(ExecutionType.MOCK, {"num_tasks": str(num_tasks)}))
    job.setup()
    return job


def mock_worker(wid, cores):
    worker = Worker(wid, None, None)
    worker.cores = cores  # hack
    return worker


@pytest.fixture
def simple_dispatcher():
    cqueue = Queue()
    workers_manager = WorkerManager(MockWorkerFactory(None))
    workers_manager.disable_monitor()
    dispatcher = FairDispatcher(cqueue, workers_manager, None)

    workers_manager.start()
    dispatcher.start()

    yield dispatcher

    dispatcher.stop()
    workers_manager.stop()


def _verify_queue_size(cqueue, size):
    return cqueue.qsize() == size


class TestSimpleDispatcher:
    def test_simple_dipstacher(self, simple_dispatcher):

        simple_dispatcher.workers_manager.register_worker(mock_worker("worker 1", 10))

        for task in mock_job_exec(2).get_tasks():
            simple_dispatcher.cqueue.put(task)

        assert wait_until(_verify_queue_size, 10, 0.5, simple_dispatcher.cqueue, 0)

    def test_dipstacher_stop(self, simple_dispatcher):

        simple_dispatcher.workers_manager.register_worker(mock_worker("worker 1", 10))

        simple_dispatcher.stop()

        for task in mock_job_exec(2).get_tasks():
            simple_dispatcher.cqueue.put(task)

        assert simple_dispatcher.cqueue.qsize() == 2

    def test_dipstacher_callback(self, simple_dispatcher):
        def callback(tid, worker):
            callback_received.append({tid: worker})

        callback_received = []

        simple_dispatcher.callback = callback
        simple_dispatcher.workers_manager.register_worker(mock_worker("worker 1", 10))

        for task in mock_job_exec(2).get_tasks():
            simple_dispatcher.cqueue.put(task)

        assert wait_until(lambda l, size: len(l) == size, 10, 0.5, callback_received, 2)


class TestFairDispatcher:
    def test_dipstacher_fairness_1(self, simple_dispatcher):

        simple_dispatcher.workers_manager.register_worker(mock_worker("worker 1", 10))
        simple_dispatcher.workers_manager.register_worker(mock_worker("worker 2", 10))

        for task in mock_job_exec(4).get_tasks():
            simple_dispatcher.cqueue.put(task)

        assert wait_until(_verify_queue_size, 10, 0.5, simple_dispatcher.cqueue, 0)

        for worker in simple_dispatcher.workers_manager.list_live_workers():
            assert len(worker.tasks) == 2

    def test_dispatcher_fairness_2(self, simple_dispatcher):
        """
        Tests fair dispatcher logic based on workers load.

        for instance:
           w1 --> 5 cores
           w2 --> 12 cores
        the dispatcher will dispatch tasks to the worker with less pending
        then the to the worker with more available cores so for (5,12) :
        (0,0) -> (0,1) -> (0,2) -> (0,3) -> (0.4) -> (0,5) -> (0,6)
        -> (0,7) -> (1,7) -> (1,8) -> (2,8) -> (2,9) -> (3,9) -> (3,10)
        -> (4,10) -> (4,11) -> (5,11) -> (5,12) -> (6,12) -> (6,13) -> (7,13)
        """
        simple_dispatcher.workers_manager.register_worker(mock_worker("worker 1", 5))
        simple_dispatcher.workers_manager.register_worker(mock_worker("worker 2", 12))

        for task in mock_job_exec(20).get_tasks():
            simple_dispatcher.cqueue.put(task)

        assert wait_until(_verify_queue_size, 10, 0.5, simple_dispatcher.cqueue, 0)

        for worker in simple_dispatcher.workers_manager.list_live_workers():
            if worker.wid == "worker 1":
                assert len(worker.tasks) == 7
            elif worker.wid == "worker 2":
                assert len(worker.tasks) == 13



import pytest

from tanit.common.model.execution_type import ExecutionType
from tanit.common.model.job import Job
from tanit.common.model.worker import Worker
from tanit.master.core.execution.execution_manager import ExecutionManager
from tanit.master.core.execution.execution_state import ExecutionState
from tanit.master.core.worker.worker_manager import WorkerManager

from ....utils.tutils import wait_until
from ..worker.mock_worker import MockWorkerFactory


def mock_worker(wid, cores):
    worker = Worker(wid, None, None)
    worker.cores = cores  # hack
    return worker


@pytest.fixture
def execution_manager():
    workers_manager = WorkerManager(MockWorkerFactory(None))
    workers_manager.register_worker(mock_worker("worker 1", 10))

    execution_manager = ExecutionManager(workers_manager)

    execution_manager.start()
    yield execution_manager
    execution_manager.stop()


def _verify_state(obj, state):
    return obj.state == state


class TestExecutionManager:
    def test_job_submit(self, execution_manager):
        execution_manager.submit_job(Job(ExecutionType.MOCK, {"num_tasks": "2"}))
        assert True

    def test_task_finish(self, execution_manager):
        job_exec = execution_manager.submit_job(
            Job(ExecutionType.MOCK, {"num_tasks": "2"})
        )

        # verify the job is in running state
        assert wait_until(
            _verify_state,
            10,
            0.5,
            execution_manager.get_job(job_exec.jid),
            ExecutionState.DISPATCHED,
        )
        for task in execution_manager.get_job(job_exec.jid).get_tasks():
            assert wait_until(_verify_state, 10, 0.5, task, ExecutionState.DISPATCHED)
            execution_manager.task_start(task.tid)

        assert execution_manager.get_job(job_exec.jid).state == ExecutionState.RUNNING

        for task in execution_manager.get_job(job_exec.jid).get_tasks():
            assert task.state == ExecutionState.RUNNING
            execution_manager.task_finish(task.tid)

        assert execution_manager.get_job(job_exec.jid).state == ExecutionState.FINISHED

        for task in execution_manager.get_job(job_exec.jid).get_tasks():
            assert task.state == ExecutionState.FINISHED

    def test_task_fail(self, execution_manager):
        # do not retry tasks on failure
        execution_manager.max_task_retries = 1

        job_exec = execution_manager.submit_job(
            Job(ExecutionType.MOCK, {"num_tasks": "2"})
        )

        # verify the job is in running state
        assert wait_until(
            _verify_state,
            10,
            0.5,
            execution_manager.get_job(job_exec.jid),
            ExecutionState.DISPATCHED,
        )

        for task in execution_manager.get_job(job_exec.jid).get_tasks():
            assert wait_until(_verify_state, 10, 0.5, task, ExecutionState.DISPATCHED)
            execution_manager.task_start(task.tid)

        assert execution_manager.get_job(job_exec.jid).state == ExecutionState.RUNNING
        for task in execution_manager.get_job(job_exec.jid).get_tasks():
            assert task.state == ExecutionState.RUNNING

        execution_manager.task_failure(
            execution_manager.get_job(job_exec.jid).get_tasks()[0].tid
        )
        for task in execution_manager.get_job(job_exec.jid).get_tasks()[1:]:
            execution_manager.task_finish(task.tid)

        assert execution_manager.get_job(job_exec.jid).state == ExecutionState.FAILED
        return job_exec

    def test_task_reset(self, execution_manager):
        job_exec = self.test_task_fail(execution_manager)

        # the failed task
        task = execution_manager.get_job(job_exec.jid).get_tasks()[0]

        # reset the failed task
        execution_manager.task_reset(task.tid)
        assert execution_manager.get_job(job_exec.jid).state == ExecutionState.RUNNING

        # wait for the task to be dispatched
        assert wait_until(_verify_state, 10, 0.5, task, ExecutionState.DISPATCHED)

        execution_manager.task_start(task.tid)
        assert task.state == ExecutionState.RUNNING
        assert execution_manager.get_job(job_exec.jid).state == ExecutionState.RUNNING

        execution_manager.task_finish(task.tid)
        assert task.state == ExecutionState.FINISHED
        assert execution_manager.get_job(job_exec.jid).state == ExecutionState.FINISHED

    def test_task_lookup(self, execution_manager):
        job_exec = execution_manager.submit_job(
            Job(ExecutionType.MOCK, {"num_tasks": "2"})
        )

        # verify the job is in running state
        assert wait_until(
            _verify_state,
            10,
            0.5,
            execution_manager.get_job(job_exec.jid),
            ExecutionState.DISPATCHED,
        )
        for task in execution_manager.get_job(job_exec.jid).get_tasks():
            assert wait_until(_verify_state, 10, 0.5, task, ExecutionState.DISPATCHED)
            execution_manager.task_start(task.tid)

        assert (
            len(
                execution_manager.get_tasks(
                    jid=job_exec.jid, states=[ExecutionState.RUNNING]
                )
            )
            == 2
        )



import json
import logging as lg
import os

_logger = lg.getLogger(__name__)


class FileSystemsConfig(object):
    config_file = "filesystems.json"
    default_conf_dir = os.path.expanduser("/etc/tanit/conf")

    def __init__(self, path=None):
        self._filesystems = {}

        self.conf_dir = path or os.getenv("TANIT_CONF_DIR", self.default_conf_dir)
        self.conf_file = os.path.join(self.conf_dir, self.config_file)

        if os.path.exists(self.conf_file):
            try:
                self.config = json.loads(open(self.conf_file).read())
            except Exception as e:
                raise TanitConfigurationException(
                    "Exception while loading configuration file %s.", self.conf_file, e
                )

            _logger.info(
                "Instantiated filesystems configuration from %s.", self.conf_file
            )
        else:
            _logger.warning(
                "Could not find filesystems configuration file, "
                + "instantiating empty configuration."
            )
            self.config = None

    def get_config(self):
        return self.config


class TanitConfigurationException(Exception):
    pass



from ...standalone.worker import LocalWorker
from .worker import RemoteThriftWorker


class WorkerFactory(object):
    def __init__(self, master):
        self.master = master

    def create_worker(self, worker):
        if worker.address is not None and worker.port is not None:
            return RemoteThriftWorker(worker.wid, worker.address, worker.port)
        else:
            return LocalWorker(worker.wid, self.master)



from .executor import Executor


class ExecutorFactory(object):
    def __init__(self, client_factory, cqueue, concurrency):
        self.client_factory = client_factory
        self.cqueue = cqueue
        self.concurrency = concurrency

    def create_executor(self, eid):
        return Executor(eid, self.cqueue, self.client_factory)



import logging as lg
import time
from threading import Thread

from six.moves.queue import Empty

from ...master.client.client import ClientType

_logger = lg.getLogger(__name__)


class Executor(Thread):
    def __init__(self, eid, cqueue, factory):
        super(Executor, self).__init__()
        self.cqueue = cqueue
        self.eid = eid
        self.stopped = False
        self.idle = True

        """
        Apparently Thrift clients are not thread safe,
        so can not use without synchronization.
        synchronizing the client calls with locks will came with performance penalty
        the best approach seems to be, using a separate thrift client
        (connection) per thread.
        """
        self.master = factory.create_client(ClientType.WORKER_SERVICE)

        self.current_task = None

    def stop(self):
        self.stopped = True

    def run(self):
        # start the client
        self.master.start()
        self._run()
        self.master.stop()

    def _run(self):
        while True:
            self.idle = True
            try:
                task_exec = self.cqueue.get(timeout=0.5)
                self.idle = False
                self.master.task_start(str(task_exec.tid))
                self.current_task = task_exec

                task_exec.run()

                self.master.task_success(str(task_exec.tid))
                self.cqueue.task_done()
            except Empty:
                if self.stopped:
                    break
                time.sleep(0.5)
            except Exception as e:
                _logger.error("Error executing task [%s]", str(task_exec.tid))
                _logger.error(e, exc_info=True)
                self.master.task_failure(str(task_exec.tid))
                self.cqueue.task_done()

    def isIdle(self):
        return self.idle



import logging as lg

from ...common.model.execution_type import ExecutionType
from ...common.model.job import Job
from ...common.model.worker import Worker
from ...thrift.master.service import ttypes
from ..core.execution.execution_state import ExecutionState

_logger = lg.getLogger(__name__)


class UserServiceHandler(object):
    def __init__(self, master):
        self.master = master

    def submit_job(self, job):
        if job.type == ttypes.JobType.COPY:
            etype = ExecutionType.COPY
        elif job.type == ttypes.JobType.UPLOAD:
            etype = ExecutionType.UPLOAD
        elif job.type == ttypes.JobType.MOCK:
            etype = ExecutionType.MOCK
        else:
            # should raise exception here
            pass

        try:
            job_exec = self.master.submit_job(Job(etype, job.params))
        except Exception as e:
            raise ttypes.JobInitializationException(str(e))
        return job_exec.jid

    def list_jobs(self):
        status = []
        for job_exec in self.master.list_jobs():
            status.append(
                ttypes.JobStatus(
                    job_exec.jid,
                    ttypes.JobState._NAMES_TO_VALUES[
                        ExecutionState._VALUES_TO_NAMES[job_exec.state]
                    ],
                    job_exec.submission_time.strftime("%Y-%m-%d %H:%M:%S"),
                    "-"
                    if job_exec.start_time is None
                    else job_exec.start_time.strftime("%Y-%m-%d %H:%M:%S"),
                    "-"
                    if job_exec.finish_time is None
                    else job_exec.finish_time.strftime("%Y-%m-%d %H:%M:%S"),
                    job_exec.execution_time_s,
                )
            )
        return status

    def job_status(self, jid):
        job_exec = self.master.get_job(jid)
        if job_exec is None:
            raise ttypes.JobNotFoundException("No such job [ %s ]" % jid)
        return ttypes.JobStatus(
            job_exec.jid,
            ttypes.JobState._NAMES_TO_VALUES[
                ExecutionState._VALUES_TO_NAMES[job_exec.state]
            ],
            job_exec.submission_time.strftime("%Y-%m-%d %H:%M:%S"),
            "-"
            if job_exec.start_time is None
            else job_exec.start_time.strftime("%Y-%m-%d %H:%M:%S"),
            "-"
            if job_exec.finish_time is None
            else job_exec.finish_time.strftime("%Y-%m-%d %H:%M:%S"),
            job_exec.execution_time_s,
        )


class WorkerServiceHandler(object):
    def __init__(self, master):
        self.master = master

    def list_workers(self):
        workers = []
        for wkr in self.master.list_workers():
            workers.append(ttypes.Worker(wkr.wid, wkr.address, wkr.port))
        return workers

    def register_heartbeat(self, worker):
        self.master.register_heartbeat(Worker(worker.wid, worker.address, worker.port))

    def register_worker(self, worker):
        self.master.register_worker(Worker(worker.wid, worker.address, worker.port))

    def unregister_worker(self, worker):
        self.master.unregister_worker(Worker(worker.wid, worker.address, worker.port))

    def register_filesystem(self, filesystem):
        self.master.register_filesystem(filesystem.name, filesystem.parameters)

    def task_start(self, tid):
        self.master.task_start(tid)

    def task_success(self, tid):
        self.master.task_success(tid)

    def task_failure(self, tid):
        self.master.task_failure(tid)

    def send_heartbeat(self, worker):
        pass


class JobNotFoundException(Exception):
    """Raised when trying to submit a task to a stopped master."""

    pass



class ExecutionType:
    COPY = 1
    UPLOAD = 2
    MOCK = 3

    _VALUES_TO_NAMES = {
        1: "COPY",
        2: "UPLOAD",
        3: "MOCK",
    }

    _NAMES_TO_VALUES = {
        "COPY": 1,
        "UPLOAD": 2,
        "MOCK": 3,
    }



#!/usr/bin/env python
# encoding: utf-8

"""tanit-server: Tanit master.

Usage:
  tanit-client [-v...] [--standalone]
  tanit-client (--version | -h)

Options:
  --version                     Show version and exit.
  -h --help                     Show help and exit.
  -v --verbose                  Enable log output. Can be specified up to three
                                times (increasing verbosity each time).
  --standalone                  Run in standalone mode.

Examples:
  tanit-client job --submit @test-job.json

"""

import logging as lg

from docopt import docopt

from ... import __version__
from .server import MasterServer


def configure_logging():
    # capture warnings issued by the warnings module
    lg.captureWarnings(True)

    logger = lg.getLogger()
    logger.setLevel(lg.DEBUG)
    lg.getLogger("requests_kerberos.kerberos_").setLevel(lg.CRITICAL)
    lg.getLogger("google.resumable_media").setLevel(lg.ERROR)
    lg.getLogger("requests").setLevel(lg.ERROR)

    # Configure stream logging if applicable
    stream_handler = lg.StreamHandler()
    stream_handler.setLevel(lg.INFO)

    fmt = "%(asctime)s\t%(name)-16s\t%(levelname)-5s\t%(message)s"
    stream_handler.setFormatter(lg.Formatter(fmt))
    logger.addHandler(stream_handler)


def main(argv=None):
    args = docopt(__doc__, argv=argv, version=__version__)

    configure_logging()

    server = MasterServer(standalone=True if args["--standalone"] else False)
    server.start()


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""tanit: python distributed data transfer tool."""

import os
import sys
import re
from setuptools import find_packages, setup

sys.path.insert(0, os.path.abspath('src'))


def _get_version():
    """Extract version from package."""
    with open('tanit/__init__.py') as reader:
        match = re.search(
            r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
            reader.read(),
            re.MULTILINE
        )
        if match:
            return match.group(1)
        else:
            raise RuntimeError('Unable to extract version.')


def _get_long_description():
    """Get README contents."""
    with open('README.rst') as reader:
        return reader.read()


setup(
    name='tanit',
    version=_get_version(),
    description=__doc__,
    keywords="thrift python tanit hdfs",
    long_description=_get_long_description(),
    author='Yassine Azzouz',
    author_email='yassine.azzouz@agmail.com',
    url='https://github.com/yassineazzouz/tanit',
    license='MIT',
    packages=find_packages(),
    classifiers=[
        'Topic :: Software Development',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*',
    install_requires=[
        'six',
        'docopt',
        'thrift>=0.10',
        'pywhdfs>=1.1.0',
        's3fs==0.2.2',
        'google-cloud-storage>=1.20.0',
        'google-auth>=1.2',
        'pyrsistent==0.15.0'
    ],
    entry_points={'console_scripts':
                      ['tanit-master = tanit.master.server.__main__:main',
                       'tanit-worker = tanit.worker.server.__main__:main',
                       'tanit-client = tanit.master.client.__main__:main']
                  },
    long_description_content_type='text/markdown'
)



import os
from threading import Thread

import pytest

from tanit.master.client.client import ClientType
from tanit.master.client.client import ThriftClientFactory
from tanit.master.config.config import MasterConfig
from tanit.master.server.server import MasterServer
from tanit.worker.server.server import WorkerServer

from ..resources import conf

config_dir = os.path.dirname(os.path.abspath(conf.__file__))


@pytest.fixture(scope="session")
def master_server():
    server = MasterServer(config=config_dir)
    server_daemon = Thread(target=server.start, args=())
    server_daemon.setDaemon(True)
    server_daemon.start()

    yield

    server.stop()
    server_daemon.join()


@pytest.fixture(scope="class")
def worker_server():
    server = WorkerServer(config=config_dir)
    server_daemon = Thread(target=server.start, args=())
    server_daemon.setDaemon(True)
    server_daemon.start()

    yield

    server.stop()
    server_daemon.join()


@pytest.fixture
def user_client():
    config = MasterConfig(path=config_dir)
    factory = ThriftClientFactory(
        config.client_service_host, config.client_service_port
    )
    client = factory.create_client(ClientType.USER_SERVICE)

    client.start()
    yield client
    client.stop()


@pytest.fixture
def worker_client():
    config = MasterConfig(path=config_dir)
    factory = ThriftClientFactory(
        config.worker_service_host, config.worker_service_port
    )
    client = factory.create_client(ClientType.WORKER_SERVICE)

    client.start()
    yield client
    client.stop()



#!/usr/bin/env python
# encoding: utf-8

import abc
import logging as lg
import time

from thrift.protocol import TBinaryProtocol
from thrift.transport import TSocket
from thrift.transport import TTransport

import six

from ...common.model.job import JobStatus
from ...common.model.worker import Worker
from ...thrift.master.service import MasterUserService
from ...thrift.master.service import MasterWorkerService
from ...thrift.master.service import ttypes

_logger = lg.getLogger(__name__)

WORKER_SERVICE_CLIENT_NAME = "worker-service"
USER_SERVICE_CLIENT_NAME = "user-service"


def connect(master_host, master_port):
    # Create Transport
    socket = TSocket.TSocket(master_host, master_port)
    transport = TTransport.TBufferedTransport(socket)

    # Connect to server
    retries = 1
    last_error = None
    while retries < 30:
        try:
            transport.open()
            break
        except TTransport.TTransportException as e:
            _logger.error(
                "Could not connect to the master server. " + "retrying in 5 seconds ..."
            )
            last_error = e
        retries += 1
        time.sleep(5.0)

    if retries == 30:
        _logger.error(
            "Could not connect to the master server after 30 retries. " + "exiting ..."
        )
        raise last_error

    return transport


@six.add_metaclass(abc.ABCMeta)
class UserServiceClientIFace(object):
    @abc.abstractmethod
    def start(self):
        raise NotImplementedError

    @abc.abstractmethod
    def list_jobs(self):
        raise NotImplementedError

    @abc.abstractmethod
    def job_status(self, jid):
        raise NotImplementedError

    @abc.abstractmethod
    def submit_job(self, job):
        raise NotImplementedError

    @abc.abstractmethod
    def stop(self):
        raise NotImplementedError


class ThriftUserServiceClient(UserServiceClientIFace):
    def __init__(self, master_host, master_port):
        self.master_host = master_host
        self.master_port = master_port

    def start(self):
        self.transport = connect(self.master_host, self.master_port)
        self.client = MasterUserService.Client(
            TBinaryProtocol.TBinaryProtocol(self.transport)
        )

    def list_jobs(self):
        jobs = []
        for job in self.client.list_jobs():
            jobs.append(
                JobStatus(
                    job.id,
                    ttypes.JobState._VALUES_TO_NAMES[job.state],
                    job.submission_time,
                    job.start_time,
                    job.finish_time,
                    job.execution_time,
                )
            )
        return jobs

    def job_status(self, jid):
        try:
            st = self.client.job_status(jid)
            return JobStatus(
                st.id,
                ttypes.JobState._VALUES_TO_NAMES[st.state],
                st.submission_time,
                st.start_time,
                st.finish_time,
                st.execution_time,
            )
        except ttypes.JobNotFoundException:
            return None

    def submit_job(self, job):
        _logger.info("Submitting new job.")
        job = ttypes.Job(job.etype, job.params)
        jid = self.client.submit_job(job)
        _logger.info("Job submitted : %s.", jid)
        return jid

    def stop(self):
        self.transport.close()


class LocalUserServiceClient(UserServiceClientIFace):
    def __init__(self, master):
        self.master = master

    def start(self):
        # do nothing
        return

    def list_jobs(self):
        return self.master.list_jobs()

    def job_status(self, jid):
        return self.master.get_job(jid)

    def submit_job(self, job):
        return self.master.submit_job()

    def stop(self):
        # do nothing
        return


@six.add_metaclass(abc.ABCMeta)
class WorkerServiceClientIFace(object):
    @abc.abstractmethod
    def start(self):
        raise NotImplementedError

    @abc.abstractmethod
    def list_workers(self):
        raise NotImplementedError

    @abc.abstractmethod
    def register_worker(self, wid, address, port):
        raise NotImplementedError

    @abc.abstractmethod
    def unregister_worker(self, wid, address, port):
        raise NotImplementedError

    @abc.abstractmethod
    def register_heartbeat(self, wid, address, port):
        raise NotImplementedError

    @abc.abstractmethod
    def task_start(self, tid):
        raise NotImplementedError

    @abc.abstractmethod
    def task_success(self, tid):
        raise NotImplementedError

    @abc.abstractmethod
    def task_failure(self, tid):
        raise NotImplementedError

    @abc.abstractmethod
    def stop(self):
        raise NotImplementedError


class ThriftWorkerServiceClient(WorkerServiceClientIFace):
    """
    Worker service Thrift client.

    Used mainly by the master to to communicate with workers.
    """

    def __init__(self, master_host, master_port):
        self.master_host = master_host
        self.master_port = master_port

    def start(self):
        self.transport = connect(self.master_host, self.master_port)
        self.client = MasterWorkerService.Client(
            TBinaryProtocol.TBinaryProtocol(self.transport)
        )

    def list_workers(self):
        wkr_list = []
        for wkr in self.client.list_workers():
            wkr_list.append(Worker(wkr.wid, wkr.address, wkr.port))
        return wkr_list

    def register_worker(self, wid, address, port):
        return self.client.register_worker(ttypes.Worker(wid, address, port))

    def unregister_worker(self, wid, address, port):
        return self.client.unregister_worker(ttypes.Worker(wid, address, port))

    def register_heartbeat(self, wid, address, port):
        self.client.register_heartbeat(ttypes.Worker(wid, address, port))

    def register_filesystem(self, name, parameters):
        self.client.register_filesystem(ttypes.FileSystem(name, parameters))

    def task_start(self, tid):
        self.client.task_start(tid)

    def task_success(self, tid):
        self.client.task_success(tid)

    def task_failure(self, tid):
        self.client.task_failure(tid)

    def stop(self):
        self.transport.close()


class LocalWorkerServiceClient(WorkerServiceClientIFace):
    """Used by local clients accessing the master worker services directly."""

    def __init__(self, master):
        self.master = master

    def start(self):
        # do nothing
        return

    def list_workers(self):
        return self.master.list_workers()

    def register_worker(self, wid, address, port):
        self.master.register_worker(Worker(self, wid, address, port))

    def unregister_worker(self, wid, address, port):
        self.master.unregister_worker(Worker(self, wid, address, port))

    def register_heartbeat(self, wid, address, port):
        self.master.register_heartbeat(Worker(self, wid, address, port))

    def task_start(self, tid):
        self.master.task_start(tid)

    def task_success(self, tid):
        self.master.task_success(tid)

    def task_failure(self, tid):
        self.master.task_failure(tid)

    def stop(self):
        # do nothing
        return


class ClientType:
    USER_SERVICE = 1
    WORKER_SERVICE = 2

    _VALUES_TO_NAMES = {
        1: "USER_SERVICE",
        2: "WORKER_SERVICE",
    }

    _NAMES_TO_VALUES = {
        "USER_SERVICE": 1,
        "WORKER_SERVICE": 2,
    }


class LocalClientFactory(object):
    def __init__(self, master):
        self.master = master

    def create_client(self, client_type):
        if client_type == ClientType.WORKER_SERVICE:
            return LocalWorkerServiceClient(self.master)

        elif client_type == ClientType.USER_SERVICE:
            return LocalUserServiceClient(self.master)
        else:
            raise NoSuchClientException(
                "No such client [ %s ]", ClientType._VALUES_TO_NAMES[client_type]
            )


class ThriftClientFactory(object):
    def __init__(self, host, port):
        self.host = host
        self.port = port

    def create_client(self, client_type):
        if client_type == ClientType.WORKER_SERVICE:
            return ThriftWorkerServiceClient(self.host, self.port)

        elif client_type == ClientType.USER_SERVICE:
            return ThriftUserServiceClient(self.host, self.port)
        else:
            raise NoSuchClientException(
                "No such client [ %s ]", ClientType._VALUES_TO_NAMES[client_type]
            )


class NoSuchClientException(Exception):
    pass



import logging as lg
from datetime import datetime
from threading import RLock

from .worker_monitor import WorkerMonitor

_logger = lg.getLogger(__name__)


class WorkerManager(object):
    """
    The WorkerManager monitor the state of workers and maintain the list of active/dead workers.
    It is not really part of the execution pipeline, its role is just to keep tack of which
    machines are active and fetch their state.
    """  # NOQA

    def __init__(self, workers_factory, config=None):
        self.worker_factory = workers_factory

        self.live_workers = []
        self.decommissioning_workers = []
        self.dead_workers = []

        self.lock = RLock()

        # monitor
        self.monitor = WorkerMonitor(self)

        self.configure(config)

    def configure(self, config):
        pass

    def start(self):
        _logger.info("Stating tanit worker manager.")
        self.monitor.start()
        _logger.info("Tanit worker manager started.")

    def stop(self):
        _logger.info("Stopping tanit worker manager.")
        self.monitor.stop()
        with self.lock:
            for wkr in self.live_workers:
                wkr.stop()
        _logger.info("Tanit worker manager stopped.")

    def disable_monitor(self):
        self.monitor.heartbeat_check_interval = -1

    def get_live_worker(self, wid):
        with self.lock:
            for wkr in self.live_workers:
                if wkr.wid == wid:
                    return wkr
            return None

    def list_live_workers(self):
        with self.lock:
            return self.live_workers

    def list_decommissioning_workers(self):
        with self.lock:
            return self.decommissioning_workers

    def register_worker(self, worker):
        with self.lock:
            for wkr in self.live_workers:
                if wkr.wid == worker.wid:
                    _logger.error("Worker [ %s ] is already registered.", worker.wid)
                    raise AlreadyRegisteredWorkerException(
                        "Worker [ %s ] is already registered.", worker.wid
                    )
                elif (
                    wkr.address == worker.address
                    and wkr.port == worker.port
                    and wkr.address is not None
                    and worker.port is not None
                ):
                    _logger.error(
                        "Worker running on address [ %s ] and port [ %s ] "
                        + "is already registered.",
                        worker.address,
                        worker.port,
                    )
                    raise AlreadyRegisteredWorkerException(
                        "Worker running on address [ %s ] and port [ %s ] "
                        + "is already registered.",
                        worker.address,
                        worker.port,
                    )
                elif wkr.address == worker.address:
                    _logger.warn(
                        "Another Worker [ %s ] is already running on [ %s ].",
                        worker.wid,
                        worker.address,
                    )

            for wkr in self.decommissioning_workers:
                if wkr.wid == worker.wid:
                    _logger.error(
                        "Worker [ %s ] is in decommissioning state, "
                        + "it can not be registered.",
                        worker.wid,
                    )
                    raise AlreadyRegisteredWorkerException(
                        "Worker [ %s ] is in decommissioning state, "
                        + "it can not be registered.",
                        worker.wid,
                    )

            remote_worker = self.worker_factory.create_worker(worker)
            remote_worker.start()
            self.live_workers.append(remote_worker)

    def register_heartbeat(self, worker):
        with self.lock:
            remote_worker = self.get_live_worker(worker.wid)
            if remote_worker is not None:
                remote_worker.last_hear_beat = datetime.now()
            else:
                raise NoSuchWorkerException("No such worker [ %s ]", worker.wid)

    def decommission_worker(self, wid):
        _logger.info("Decommisionning Worker [ %s ]", wid)
        with self.lock:
            remote_worker = self.get_live_worker(wid)
            if remote_worker is not None:
                self.live_workers.remove(remote_worker)
                self.decommissioning_workers.append(remote_worker)
            else:
                raise NoSuchWorkerException("No such worker [ %s ]", wid)
        _logger.info("Worker [ %s ] Decommisionned", wid)

    def on_worker_decommissioned(self, wid):
        _logger.info("Deleting Worker [ %s ]", wid)
        with self.lock:
            for wkr in self.decommissioning_workers:
                if wkr.wid == wid:
                    wkr.stop()
                    self.decommissioning_workers.remove(wkr)
                    self.dead_workers.append(wkr)
                    break
        _logger.info("Worker [ %s ] Deleted", wid)


class IllegalWorkerStateException(Exception):
    pass


class NoSuchWorkerException(Exception):
    pass


class AlreadyRegisteredWorkerException(Exception):
    pass



import abc
import logging as lg
from datetime import datetime

import six

from ....worker.client.client import WorkerClient

_logger = lg.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class WorkerIFace(object):
    def __init__(self, wid, address, port):
        self.wid = wid
        self.address = address
        self.port = port
        self.last_hear_beat = datetime.now()

        self.stopped = True

    def start(self):
        self.stopped = False

    def stop(self):
        self.stopped = True

    @abc.abstractmethod
    def submit(self, task):
        pass

    @abc.abstractmethod
    def register_filesystem(self, name, filesystem):
        pass

    @abc.abstractmethod
    def status(self):
        pass


class RemoteThriftWorker(WorkerIFace):
    def __init__(self, wid, address, port):
        super(RemoteThriftWorker, self).__init__(wid, address, port)
        self.client = WorkerClient(address, port)

    def start(self):
        super(RemoteThriftWorker, self).start()
        self.client.start()

    def stop(self):
        super(RemoteThriftWorker, self).stop()
        self.client.stop()

    def submit(self, task_exec):
        if not self.stopped:
            self.client.submit(task_exec.tid, task_exec.etype, task_exec.params)
        else:
            raise WorkerStoppedException(
                "Can not submit task [ %s ] to [ %s ] : worker stopped.",
                task_exec.tid,
                self.wid,
            )

    def register_filesystem(self, name, filesystem):
        if not self.stopped:
            self.client.register_filesystem(name, filesystem)
        else:
            raise WorkerStoppedException(
                "Can not register filesystem [ %s ] : worker stopped." % name
            )

    def status(self):
        return self.client.worker_status()


class WorkerStoppedException(Exception):
    """Raised when trying to submit a task to a stopped worker."""

    pass



import random
import string

from tanit.filesystem.ioutils import ChunkFileReader


class MockDataGenerator:
    def __init__(self, size):
        self.size = size
        self.pos = 0

    def get_random_string(self, length):
        letters = string.ascii_lowercase
        return "".join(random.choice(letters) for i in range(length))

    def read(self, length):
        if self.size == self.pos:
            return None
        else:
            self.pos += length
            return self.get_random_string(min(length, self.size - self.pos))


class TestReader:
    def test_chunk_reader(self):
        reader = ChunkFileReader(MockDataGenerator(32), 16)
        for data in reader:
            assert len(data) == 16

        reader = ChunkFileReader(MockDataGenerator(34), 16)
        for data in reader:
            last_data = data
        assert len(last_data) == 2



from thrift.protocol import TBinaryProtocol

# Thrift files
from thrift.transport import TSocket
from thrift.transport import TTransport

from ...common.model.worker import WorkerStatus
from ...thrift.worker.service import WorkerService
from ...thrift.worker.service.ttypes import FileSystem
from ...thrift.worker.service.ttypes import Task


class WorkerClient(object):
    def __init__(self, host, port):
        self.host = host
        self.port = port
        # Init thrift connection and protocol handlers
        socket = TSocket.TSocket(host, port)
        self.transport = TTransport.TBufferedTransport(socket)
        protocol = TBinaryProtocol.TBinaryProtocol(self.transport)

        # Set client to our Example
        self.client = WorkerService.Client(protocol)

    def start(self):
        # Connect to server
        self.transport.open()

    def submit(self, tid, etype, params):
        self.client.submit(Task(tid, etype, params))

    def register_filesystem(self, name, parameters):
        self.client.register_filesystem(FileSystem(name, parameters))

    def worker_status(self):
        status = self.client.worker_status()
        return WorkerStatus(
            status.wid, status.running, status.pending, status.available
        )

    def stop(self):
        self.transport.close()



import abc
import logging as lg
import os.path as osp
import time
from threading import Lock

import six

from ....common.core.exception import TanitError
from ....common.utils.utils import str2bool
from ....filesystem.filesystem_factory import FileSystemFactory

_logger = lg.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class TaskExecution(object):
    def __init__(self, tid, params):
        self.tid = tid
        self.params = params

        self.initialize(params)

    @abc.abstractmethod
    def initialize(self, params):
        return

    @abc.abstractmethod
    def run(self):
        return


class UploadTaskExecution(TaskExecution):
    # not implemented yet
    pass


class MockTaskExecution(TaskExecution):
    def initialize(self, params):
        self.fail = str2bool(params["fail"]) if "fail" in params else False
        self.sleep = float(params["sleep"]) if "sleep" in params else 2.0

    def run(self):
        time.sleep(self.sleep)
        if self.fail:
            raise TaskFailedException("Task forced failure.")
        return


class CopyTaskExecution(TaskExecution):
    def initialize(self, params):

        if "src" in params:
            self.src = params["src"]
        else:
            raise TaskInitializationException(
                "missing required copy job parameter 'src'"
            )

        if "dst" in params:
            self.dst = params["dst"]
        else:
            raise TaskInitializationException(
                "missing required copy job parameter 'dst'"
            )

        if "src_path" in params:
            self.src_path = params["src_path"]
        else:
            raise TaskInitializationException(
                "missing required copy job parameter 'src_path'"
            )

        if "dest_path" in params:
            self.dest_path = params["dest_path"]
        else:
            raise TaskInitializationException(
                "missing required copy job parameter 'dest_path'"
            )

        self.force = str2bool(params["force"]) if "force" in params else False
        self.overwrite = (
            str2bool(params["overwrite"]) if "overwrite" in params else True
        )
        self.checksum = str2bool(params["checksum"]) if "checksum" in params else False
        self.chunk_size = int(params["chunk_size"]) if "chunk_size" in params else 65536
        self.buffer_size = (
            int(params["buffer_size"]) if "buffer_size" in params else 65536
        )

    def run(self):
        # Can cache the clients in the engine
        src = FileSystemFactory.getInstance().get_filesystem(self.src)
        dst = FileSystemFactory.getInstance().get_filesystem(self.dst)

        _src_path = self.src_path
        _dst_path = self.dest_path
        _tmp_path = ""

        lock = Lock()

        skip = False

        dst_st = dst.status(_dst_path, strict=False)

        if dst_st is None:
            # destination does not exist
            _tmp_path = _dst_path
        else:
            if not self.force:
                if self.checksum is True:
                    _src_path_checksum = src.checksum(_src_path)
                    _dst_path_checksum = dst.checksum(_dst_path)
                    if _src_path_checksum != _dst_path_checksum:
                        _logger.info(
                            "source and destination files does not seems "
                            + "to have the same checksum value, copying..."
                        )
                        skip = False
                    else:
                        _logger.info(
                            "source %r and destination %r seems to be identical, "
                            + "skipping.",
                            _src_path,
                            _dst_path,
                        )
                        skip = True
                else:
                    # By default check only the file sizes
                    if (
                        src.status(_src_path)["length"]
                        == dst.status(_dst_path)["length"]
                    ):
                        _logger.info(
                            "source %r and destination %r seems to have the same size, "
                            + "skipping.",
                            _src_path,
                            _dst_path,
                        )
                        skip = True
                    else:
                        _logger.info(
                            "source and destination files does not seems "
                            + "to have the same size, copying."
                        )
                        skip = False
            else:
                # force will always force the copy
                skip = False

            # destination exist
            if not self.overwrite and not skip:
                raise TanitError(
                    "Destination file exist and Missing overwrite parameter."
                )
            _tmp_path = "%s.temp-%s" % (_dst_path, int(time.time()))

        if not skip:
            # Prevent race condition when creating directories
            with lock:
                if dst.status(osp.dirname(_tmp_path), strict=False) is None:
                    _logger.debug(
                        "Parent directory %r does not exist, creating recursively.",
                        osp.dirname(_tmp_path),
                    )
                    curpath = ""
                    root_dir = None
                    for dirname in osp.dirname(_tmp_path).strip("/").split("/"):
                        curpath = "/".join([curpath, dirname])
                        if dst.status(curpath, strict=False) is None:
                            if root_dir is not None:
                                root_dir = curpath
                            dst.mkdir(curpath)

            _logger.info("Copying %r to %r.", _src_path, _tmp_path)

            kwargs = {}

            with src.read(
                _src_path, chunk_size=self.chunk_size, buffer_size=self.buffer_size
            ) as _reader:
                dst.write(_tmp_path, _reader, buffer_size=self.buffer_size, **kwargs)

            if _tmp_path != _dst_path:
                _logger.info(
                    "Copy of %r complete. Moving from %r to %r.",
                    _src_path,
                    _tmp_path,
                    _dst_path,
                )
                dst.delete(_dst_path)
                dst.rename(_tmp_path, _dst_path)
            else:
                _logger.info("Copy of %r to %r complete.", _src_path, _dst_path)

            return {"status": "copied", "src_path": _src_path, "dest_path": _dst_path}
        else:
            _logger.info("Skipping copy %r to %r.", _src_path, _dst_path)
            return {"status": "skipped", "src_path": _src_path, "dest_path": _dst_path}


class TaskInitializationException(Exception):
    pass


class TaskFailedException(Exception):
    pass



import logging as lg
import os
import re
import time

import s3fs

from ...common.utils.glob import iglob
from ..filesystem import IFileSystem
from ..ioutils import FileSystemError

try:
    FileNotFoundError = FileNotFoundError
except NameError:
    FileNotFoundError = s3fs.core.FileNotFoundError

_logger = lg.getLogger(__name__)


# TODO : change S3FileSystem to boto3 instead of s3fs
class S3FileSystem(IFileSystem):
    def __init__(self, bucket_name, **params):
        self.s3 = s3fs.S3FileSystem(**params)
        self.bucket_name = bucket_name

    def resolve_path(self, path):
        if path == self.bucket_name:
            rpath = self.bucket_name + "/"
        elif not path.startswith(self.bucket_name + "/"):
            if path.startswith("/"):
                rpath = self.bucket_name + path
            else:
                rpath = self.bucket_name + "/" + path
        else:
            rpath = path
        return os.path.normpath(rpath)

    def format_path(self, path):
        if path.startswith(self.bucket_name + "/"):
            return re.sub("^%s" % self.bucket_name, "", path)
        else:
            return path

    def _status(self, path, strict=True):
        def _datetime2timestamp(dt):
            try:
                int(dt.timestamp() * 1000)
            except AttributeError:
                # python 2.7
                int(time.mktime(dt.timetuple()) * 1000)

        _logger.debug("Fetching status for %r.", path)
        rpath = self.resolve_path(path)
        if not self.s3.exists(rpath):
            if not strict:
                return None
            else:
                raise FileSystemError("%r does not exist.", rpath)
        else:
            if rpath == self.bucket_name:
                # this is equivalent to root '/'
                info = {"Key": self.bucket_name, "Size": 0, "StorageClass": "DIRECTORY"}
            else:
                try:
                    info = self.s3.info(rpath)
                except FileNotFoundError:
                    # ugly workaround
                    self.s3.ls(os.path.dirname(rpath), refresh=True)
                    info = self.s3.info(rpath)
            return {
                "fileId": str(info["Key"]),
                "length": str(info["Size"])
                if str(info["StorageClass"]).upper() == "STANDARD"
                else 0,
                "type": "FILE"
                if str(info["StorageClass"]).upper() == "STANDARD"
                else "DIRECTORY",
                "modificationTime": str(_datetime2timestamp(info["LastModified"]))
                if info["StorageClass"] == "STANDARD"
                else None,
            }

    def _list(self, path, status=False, glob=False):
        if not glob:
            files = list(
                filter(
                    None,
                    [
                        self.format_path(os.path.basename(f))
                        for f in self.s3.ls(path, refresh=True)
                    ],
                )
            )
            if status:
                return [(f, self.status(os.path.join(path, f))) for f in files]
            else:
                return files
        else:
            files = [self.format_path(i_file) for i_file in iglob(self, path)]
            if status:
                return [(f, self.status(f)) for f in files]
            else:
                return files

    def _delete(self, path, recursive=False):
        self.s3.rm(path, recursive=recursive)

    def _copy(self, src_path, dst_path):
        self.s3.copy(src_path, dst_path)

    def _set_owner(self, path, owner=None, group=None):
        raise NotImplementedError

    def _set_permission(self, path, permission):
        self.s3.chmod(path, permission)

    def _mkdir(self, path, permission=None):
        self.s3.mkdir(path, permission)

    def _open(self, path, mode, buffer_size=-1, encoding=None, **kwargs):
        # omit buffer_size, s3 lib use whole block buffering
        return self.s3.open(path, mode=mode, encoding=encoding)



import logging as lg
import time

_logger = lg.getLogger(__name__)


class ExecutorPool(object):
    def __init__(self, wid, factory, cqueue, concurrency):
        self.cqueue = cqueue
        self.concurrency = concurrency
        self.wid = wid
        self.executors = []
        self.stopped = True

        for i in range(0, concurrency):
            self.executors.append(factory.create_executor("%s-executor-%s" % (wid, i)))

    def start(self):
        _logger.info("Starting Tanit worker executor pool [ %s ].", self.wid)
        self.stopped = False
        for executor in self.executors:
            _logger.info("Starting executor [ %s ]", executor.eid)
            executor.start()
            _logger.info("Executor [ %s ] started", executor.eid)
        _logger.info(
            "Tanit worker executor pool [ %s ] started, "
            + "using %s concurrent executors",
            self.wid,
            self.concurrency,
        )

    def stop(self):
        _logger.info("Stopping Tanit worker executor pool [ %s ].", self.wid)

        self.stopped = True
        for executor in self.executors:
            _logger.info("Stopping executor [ %s ]", executor.eid)
            executor.stop()
        for executor in self.executors:
            executor.join()
            _logger.info("Executor [ %s ] stopped", executor.eid)

        _logger.info("Tanit worker executor pool [ %s ] stopped.", self.wid)

    def num_running(self):
        return self.concurrency - self.num_available()

    def num_pending(self):
        return self.cqueue.qsize()

    def num_available(self):
        # if self.num_pending > 0:
        #    return self.concurrency
        idl = 0
        for executor in self.executors:
            if executor.isIdle():
                idl += 1
        return idl

    def log_stats(self):
        while not self.stopped:
            _logger.info(
                "worker starts { running : %s, pending : %s, available: %s}",
                self.num_running(),
                self.num_pending(),
                self.num_available(),
            )
            if self.executors[0].current_task is not None:
                _logger.info(
                    "task %s , %s",
                    self.executors[0].current_task.src_path,
                    self.executors[0].current_task.dest_path,
                )
            time.sleep(2)



from ....common.model.execution_type import ExecutionType
from .execution_job import CopyJobExecution
from .execution_job import MockJobExecution
from .execution_job import UploadJobExecution


class JobFactory(object):
    def create_job(self, job):
        if job.etype == ExecutionType.COPY:
            return CopyJobExecution(job.params)
        elif job.etype == ExecutionType.UPLOAD:
            return UploadJobExecution(job.params)
        elif job.etype == ExecutionType.MOCK:
            return MockJobExecution(job.params)
        else:
            raise UnknownJobTypeException("Unknown job type %s" % job.etype)


class UnknownJobTypeException(Exception):
    pass



#
# Autogenerated by Thrift Compiler (0.13.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class TaskType(object):
    COPY = 1
    UPLOAD = 2
    MOCK = 3

    _VALUES_TO_NAMES = {
        1: "COPY",
        2: "UPLOAD",
        3: "MOCK",
    }

    _NAMES_TO_VALUES = {
        "COPY": 1,
        "UPLOAD": 2,
        "MOCK": 3,
    }


class WorkerStatus(object):
    """
    Attributes:
     - wid
     - running
     - pending
     - available

    """


    def __init__(self, wid=None, running=None, pending=None, available=None,):
        self.wid = wid
        self.running = running
        self.pending = pending
        self.available = available

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.wid = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I32:
                    self.running = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I32:
                    self.pending = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.I32:
                    self.available = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('WorkerStatus')
        if self.wid is not None:
            oprot.writeFieldBegin('wid', TType.STRING, 1)
            oprot.writeString(self.wid.encode('utf-8') if sys.version_info[0] == 2 else self.wid)
            oprot.writeFieldEnd()
        if self.running is not None:
            oprot.writeFieldBegin('running', TType.I32, 2)
            oprot.writeI32(self.running)
            oprot.writeFieldEnd()
        if self.pending is not None:
            oprot.writeFieldBegin('pending', TType.I32, 3)
            oprot.writeI32(self.pending)
            oprot.writeFieldEnd()
        if self.available is not None:
            oprot.writeFieldBegin('available', TType.I32, 4)
            oprot.writeI32(self.available)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Task(object):
    """
    Attributes:
     - tid
     - type
     - params

    """


    def __init__(self, tid=None, type=None, params=None,):
        self.tid = tid
        self.type = type
        self.params = params

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.tid = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I32:
                    self.type = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.MAP:
                    self.params = {}
                    (_ktype1, _vtype2, _size0) = iprot.readMapBegin()
                    for _i4 in range(_size0):
                        _key5 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        _val6 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.params[_key5] = _val6
                    iprot.readMapEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Task')
        if self.tid is not None:
            oprot.writeFieldBegin('tid', TType.STRING, 1)
            oprot.writeString(self.tid.encode('utf-8') if sys.version_info[0] == 2 else self.tid)
            oprot.writeFieldEnd()
        if self.type is not None:
            oprot.writeFieldBegin('type', TType.I32, 2)
            oprot.writeI32(self.type)
            oprot.writeFieldEnd()
        if self.params is not None:
            oprot.writeFieldBegin('params', TType.MAP, 3)
            oprot.writeMapBegin(TType.STRING, TType.STRING, len(self.params))
            for kiter7, viter8 in self.params.items():
                oprot.writeString(kiter7.encode('utf-8') if sys.version_info[0] == 2 else kiter7)
                oprot.writeString(viter8.encode('utf-8') if sys.version_info[0] == 2 else viter8)
            oprot.writeMapEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.tid is None:
            raise TProtocolException(message='Required field tid is unset!')
        if self.type is None:
            raise TProtocolException(message='Required field type is unset!')
        if self.params is None:
            raise TProtocolException(message='Required field params is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class FileSystem(object):
    """
    Attributes:
     - name
     - parameters

    """


    def __init__(self, name=None, parameters=None,):
        self.name = name
        self.parameters = parameters

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.name = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.MAP:
                    self.parameters = {}
                    (_ktype10, _vtype11, _size9) = iprot.readMapBegin()
                    for _i13 in range(_size9):
                        _key14 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        _val15 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.parameters[_key14] = _val15
                    iprot.readMapEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('FileSystem')
        if self.name is not None:
            oprot.writeFieldBegin('name', TType.STRING, 1)
            oprot.writeString(self.name.encode('utf-8') if sys.version_info[0] == 2 else self.name)
            oprot.writeFieldEnd()
        if self.parameters is not None:
            oprot.writeFieldBegin('parameters', TType.MAP, 2)
            oprot.writeMapBegin(TType.STRING, TType.STRING, len(self.parameters))
            for kiter16, viter17 in self.parameters.items():
                oprot.writeString(kiter16.encode('utf-8') if sys.version_info[0] == 2 else kiter16)
                oprot.writeString(viter17.encode('utf-8') if sys.version_info[0] == 2 else viter17)
            oprot.writeMapEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.name is None:
            raise TProtocolException(message='Required field name is unset!')
        if self.parameters is None:
            raise TProtocolException(message='Required field parameters is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(WorkerStatus)
WorkerStatus.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'wid', 'UTF8', None, ),  # 1
    (2, TType.I32, 'running', None, None, ),  # 2
    (3, TType.I32, 'pending', None, None, ),  # 3
    (4, TType.I32, 'available', None, None, ),  # 4
)
all_structs.append(Task)
Task.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'tid', 'UTF8', None, ),  # 1
    (2, TType.I32, 'type', None, None, ),  # 2
    (3, TType.MAP, 'params', (TType.STRING, 'UTF8', TType.STRING, 'UTF8', False), None, ),  # 3
)
all_structs.append(FileSystem)
FileSystem.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'name', 'UTF8', None, ),  # 1
    (2, TType.MAP, 'parameters', (TType.STRING, 'UTF8', TType.STRING, 'UTF8', False), None, ),  # 2
)
fix_spec(all_structs)
del all_structs



from ....common.model.execution_type import ExecutionType
from .execution_task import CopyTaskExecution
from .execution_task import MockTaskExecution
from .execution_task import UploadTaskExecution


class TaskFactory(object):
    def create_task(self, task):
        if task.etype == ExecutionType.COPY:
            return CopyTaskExecution(task.tid, task.params)
        elif task.etype == ExecutionType.UPLOAD:
            return UploadTaskExecution(task.tid, task.params)
        elif task.etype == ExecutionType.MOCK:
            return MockTaskExecution(task.tid, task.params)
        else:
            raise UnknownTaskTypeException("Unknown task type %s" % task.etype)


class UnknownTaskTypeException(Exception):
    pass
